@@ -24,7 +24,6 @@ import (
 	"log"
 	"os"
 	"path"
-	"path/filepath"
 	"strings"
 )
 
@@ -45,13 +44,10 @@ func init() {
 }
 
 func redoDirClean(root, what string) error {
-	root, err := filepath.Abs(root)
-	if err != nil {
-		panic(err)
-	}
+	root = mustAbs(root)
 	dir, err := os.Open(root)
 	if err != nil {
-		return err
+		return ErrLine(err)
 	}
 	defer dir.Close()
 	for {
@@ -60,7 +56,7 @@ func redoDirClean(root, what string) error {
 			if err == io.EOF {
 				break
 			}
-			return err
+			return ErrLine(err)
 		}
 		var pth string
 		for _, entry := range entries {
@@ -72,7 +68,7 @@ func redoDirClean(root, what string) error {
 					fmt.Println(pth)
 					if !*DryRun {
 						if err = os.Remove(pth); err != nil {
-							return err
+							return ErrLine(err)
 						}
 					}
 				}
@@ -81,7 +77,7 @@ func redoDirClean(root, what string) error {
 					fmt.Println(pth)
 					if !*DryRun {
 						if err = os.Remove(pth); err != nil {
-							return err
+							return ErrLine(err)
 						}
 					}
 				}
@@ -90,12 +86,12 @@ func redoDirClean(root, what string) error {
 					fmt.Println(pth)
 					if !*DryRun {
 						if err = os.Remove(pth); err != nil {
-							return err
+							return ErrLine(err)
 						}
 					}
 				}
 			default:
-				log.Fatalln("unknown cleanup target")
+				log.Fatal("unknown cleanup target")
 			}
 		}
 	}
@@ -103,13 +99,10 @@ func redoDirClean(root, what string) error {
 }
 
 func cleanupWalker(root, what string) error {
-	root, err := filepath.Abs(root)
-	if err != nil {
-		panic(err)
-	}
+	root = mustAbs(root)
 	dir, err := os.Open(root)
 	if err != nil {
-		return err
+		return ErrLine(err)
 	}
 	defer dir.Close()
 	for {
@@ -118,7 +111,7 @@ func cleanupWalker(root, what string) error {
 			if err == io.EOF {
 				break
 			}
-			return err
+			return ErrLine(err)
 		}
 		for _, entry := range entries {
 			pth := path.Join(root, entry.Name())
@@ -128,7 +121,7 @@ func cleanupWalker(root, what string) error {
 					if what == CleanupFull {
 						fmt.Println(pthRel)
 						if !*DryRun {
-							err = os.RemoveAll(pth)
+							err = ErrLine(os.RemoveAll(pth))
 						}
 					} else {
 						err = redoDirClean(pth, what)
@@ -137,7 +130,7 @@ func cleanupWalker(root, what string) error {
 					strings.HasPrefix(entry.Name(), TmpPrefix) {
 					fmt.Println(pthRel)
 					if !*DryRun {
-						err = os.RemoveAll(pth)
+						err = ErrLine(os.RemoveAll(pth))
 					}
 				} else {
 					err = cleanupWalker(pth, what)
@@ -152,7 +145,7 @@ func cleanupWalker(root, what string) error {
 				fmt.Println(pthRel)
 				if !*DryRun {
 					if err = os.Remove(pth); err != nil {
-						return err
+						return ErrLine(err)
 					}
 				}
 			}
