@@ -371,7 +371,7 @@ func runScript(tgt *Tgt, errs chan error, forced, traced bool) error {
 	var cmdName string
 	var args []string
 	if err = unix.Access(doFile.rel, unix.X_OK); err == nil {
-		cmdName = doFileT
+		cmdName = doFile.a
 		args = make([]string, 0, 3)
 	} else {
 		cmdName = "/bin/sh"
